# -*- coding: utf-8 -*-
"""
Updater for nsupdate.info dynamic dns service (which is dyndns2 compatible,
so this module is mostly for the sake of a different _updateurl and docs).

To avoid SSL certificate issues when using https, using Python >= 3.2 is
strongly recommended (SSL with SNI is painful with requests on Python 2.x).

Updating IPv4 (uses web based ip autodetection):
dyndnsc --hostname test.nsupdate.info \
        --userid   test.nsupdate.info --password xxxxxxxxxx \
        --protocol nsupdate \
        --loop --sleeptime 300 \
        --method=webcheck

Updating IPv6 (uses interface based ip detection):
dyndnsc --hostname test.nsupdate.info \
        --userid   test.nsupdate.info --password xxxxxxxxxx \
        --protocol nsupdate \
        --loop --sleeptime 300 \
        --method=Iface,netmask:2001:470:1234:5678::/64,iface:eth0,family:INET6
"""

from .base import UpdateProtocol


class UpdateProtocolNsUpdate(UpdateProtocol):
    """Protocol handler for nsupdate.info"""

    _updateurl = "https://nsupdate.info/nic/update"

    def __init__(self, options):
        self.theip = None
        self.hostname = options['hostname']
        self.userid = options['userid']
        self.password = options['password']

        super(UpdateProtocolNsUpdate, self).__init__()

    @staticmethod
    def configuration_key():
        return "nsupdate"

    def update(self, ip):
        self.theip = ip
        return self.protocol()



# -*- coding: utf-8 -*-
"""
observer/subject implementation
"""

import logging

log = logging.getLogger(__name__)


class Subject(object):
    """
    Subject -> dispatches messages to interested callables
    """
    def __init__(self):
        self._observers = {}

    def register_observer(self, observer, events=None):
        """
        register a listener function

        Parameters
        -----------
        observer : external listener function
        events  : tuple or list of relevant events (default=None)
        """
        if events is not None and type(events) not in (tuple, list):
            events = (events,)

        if observer in self._observers:
            log.warn("Observer '%r' already registered, overwriting for events"
                     " %r", observer, events)
        self._observers[observer] = events

    def notify_observers(self, event=None, msg=None):
        """notify observers """
        for observer, events in list(self._observers.items()):
            #log.debug("trying to notify the observer")
            if events is None or event is None or event in events:
                try:
                    observer(self, event, msg)
                except (Exception,) as ex:  # pylint: disable=W0703
                    self.unregister_observer(observer)
                    errmsg = "Exception in message dispatch: Handler '{0}' unregistered for event '{1}'  ".format(observer.__class__.__name__, event)
                    log.error(errmsg, exc_info=ex)

    def unregister_observer(self, observer):
        """ unregister observer function """
        del self._observers[observer]



# -*- coding: utf-8 -*-

from .base import IPDetector


class IPDetector_Command(IPDetector):
    """IPDetector to detect IP adress executing shell command/script"""
    def __init__(self, options):
        """
        Constructor
        @param options: dictionary

        available options:

        command: shell command that writes IP address to STDOUT
        """
        self.opts = {'command': ''}
        for k in options.keys():
            self.opts[k] = options[k]
        super(IPDetector_Command, self).__init__()

    @staticmethod
    def getName():
        return "command"

    def can_detect_offline(self):
        """Returns false, as this detector possibly generates network traffic"""
        return False

    def setHostname(self, hostname):
        self.hostname = hostname

    def detect(self):
        import sys
        if sys.version_info >= (3, 0):
            import subprocess
        else:
            import commands as subprocess
        try:
            theip = subprocess.getoutput(self.opts['command'])
        except:
            theip = None
        self.set_current_value(theip)
        return theip



# -*- coding: utf-8 -*-

from .base import UpdateProtocol


class UpdateProtocolDyndns(UpdateProtocol):
    """Protocol handler for dyndns.com"""

    _updateurl = "https://members.dyndns.org/nic/update"

    def __init__(self, options):
        self.hostname = options['hostname']
        self.userid = options['userid']
        self.password = options['password']

        super(UpdateProtocolDyndns, self).__init__()

    @staticmethod
    def configuration_key():
        return "dyndns"

    def update(self, ip):
        self.theip = ip
        return self.protocol()



# -*- coding: utf-8 -*-

from .base import UpdateProtocol


class UpdateProtocolNoip(UpdateProtocol):
    """Protocol handler for www.noip.com"""

    _updateurl = "https://dynupdate.no-ip.com/nic/update"

    def __init__(self, options):
        self.theip = None
        self.hostname = options['hostname']
        self.userid = options['userid']
        self.password = options['password']

        super(UpdateProtocolNoip, self).__init__()

    @staticmethod
    def configuration_key():
        return "noip"

    def update(self, ip):
        self.theip = ip
        return self.protocol()



# -*- coding: utf-8 -*-


def detector_classes():
    from . import builtin
    return builtin.plugins


def get_detector_class(detector="webcheck"):
    detector = detector.lower()
    for cls in detector_classes():
        if cls.getName().lower() == detector:
            return cls
    raise KeyError("No IPDetector plugin named '%s' could be found" % detector)



# -*- coding: utf-8 -*-

from .dyndns import UpdateProtocolDyndns
from .dummy import UpdateProtocolDummy
from .noip import UpdateProtocolNoip
from .nsupdate_info import UpdateProtocolNsUpdate

import logging

log = logging.getLogger(__name__)


def updaterClasses():
    from .base import UpdateProtocol
    return [cls for cls in UpdateProtocol.__subclasses__()]  # + updaterClasses_external()


def updaterClasses_external():
    '''
    Tentative support for external plugins
    '''
    from pkg_resources import iter_entry_points
    for ep in iter_entry_points(group='dyndnsc.updater_builtin', name=None):
        log.warn("%s, %s", ep, type(ep))
        log.warn("%s %s %s %s %s", ep.name, ep.dist, ep.module_name, ep.attrs, ep.extras)
    return [ep.load() for ep in iter_entry_points(group='dyndnsc.updater', name=None)]


def getUpdaterClass(protoname='dyndns'):
    """factory method to get the correct protocol Handler given its name"""
    for cls in updaterClasses():
        if cls.configuration_key() == protoname:
            return cls
    raise KeyError("No UpdateProtocol registered for '%s'", protoname)



# -*- coding: utf-8 -*-

import socket

from .base import IPDetector


class IPDetector_DNS(IPDetector):
    """Class to resolve a hostname using socket.getaddrinfo()"""
    def __init__(self, hostname):
        self._hostname = hostname
        super(IPDetector_DNS, self).__init__()

    @staticmethod
    def getName():
        return "dns"

    def can_detect_offline(self):
        """Returns false, as this detector generates dns traffic"""
        return False

    def detect(self):
        try:
            theip = socket.getaddrinfo(self._hostname, None)[0][4][0]
        except:
            # logging.debug("WARN: dns is None")
            theip = None
        # logging.debug("dnsdetect for %s: %s", self.hostname, ip)
        self.set_current_value(theip)
        return theip



# -*- coding: utf-8 -*-

"""
All built-in detector plugins are listed here and will be dynamically imported
on importing this module. If importing a plugin fails, it will be ignored.
"""

from importlib import import_module
from warnings import warn

_builtins = (
    ('dyndnsc.detector.command', 'IPDetector_Command'),
    ('dyndnsc.detector.dns', 'IPDetector_DNS'),
    ('dyndnsc.detector.iface', 'IPDetector_Iface'),
    ('dyndnsc.detector.rand', 'IPDetector_Random'),
    ('dyndnsc.detector.teredo', 'IPDetector_Teredo'),
    ('dyndnsc.detector.webcheck', 'IPDetector_WebCheck'),
    )


def _load_plugin(module, cls):
    '''
    Try to load one plugin, return None if it failed to load
    :param module: module name
    :param cls: class name
    '''
    try:
        plugmod = import_module(module)
    except Exception as exc:
        warn("Importing built-in plugin %s.%s raised an exception: %r" %
             (module, cls, repr(exc)), ImportWarning)
        return None
    else:
        return getattr(plugmod, cls)


def _iload_plugins(builtins):
    '''
    return a generator for all 'importable' built-in plugins
    :param builtins: set of tuples(modulename, classname)
    '''
    for m, c in builtins:
        plugcls = _load_plugin(m, c)
        if plugcls is not None:
            yield plugcls

plugins = set((_plugcls for _plugcls in _iload_plugins(_builtins)))



# -*- coding: utf-8 -*-

import logging

import IPy
import netifaces

from .base import IPDetector

log = logging.getLogger(__name__)


class IPDetector_Iface(IPDetector):
    """IPDetector to detect any ip address of a local interface.
    """
    def __init__(self, options=None):
        """
        Constructor
        @param options: dictionary

        available options:

        iface: name of interface (default: en0)
        family: IP address family (default: INET, possible: INET6)
        netmask: netmask to be matched if multiple IPs on interface (default:
        none (match all)", example for teredo: "2001:0000::/32")
        """
        if options is None:
            options = {}
        self.opts = {'iface': 'en0', 'family': "INET", "netmask": None}
        for k in options.keys():
            log.debug("%s explicitly got option: %s -> %s",
                      self.__class__.__name__, k, options[k])
            self.opts[k] = options[k]
        super(IPDetector_Iface, self).__init__()

    @staticmethod
    def getName():
        return "iface"

    def can_detect_offline(self):
        """Returns true, as this detector only queries local data"""
        return True

    def _detect(self):
        """uses the netifaces module to detect ifconfig information"""
        if (not 'netmask' in vars(self)):
            if (not self.opts['netmask'] is None):  # if a netmask was given
                try:
                    self.netmask = IPy.IP(self.opts['netmask'])
                except (TypeError, ValueError) as e:
                    # TODO: this is a potential trust issue, because if we don't
                    # fail here, we might end up sending an IP to the outside
                    # world that should be hidden (because in a "private"
                    # netmask)
                    log.error("Choked while parsing netmask '%s'",
                              self.opts["netmask"], exc_info=e)
                    self.netmask = None

            else:
                self.netmask = None
        theip = None
        try:
            if self.opts['family'] == 'INET6':
                addrlist = netifaces.ifaddresses(self.opts['iface'])[netifaces.AF_INET6]
            else:
                addrlist = netifaces.ifaddresses(self.opts['iface'])[netifaces.AF_INET]
        except Exception as e:
            log.error("netifaces choked while trying to get inet6 interface"
                      " information for interface '%s'", self.opts['iface'],
                      exc_info=e)
        else:  # now we have a list of addresses as returned by netifaces
            for pair in addrlist:
                try:
                    detip = IPy.IP(pair['addr'])
                except (TypeError, ValueError) as e:
                    log.debug("Found invalid IP '%s' on interface '%s'!?",
                              pair['addr'], self.opts['iface'])
                    continue
                if (not self.netmask is None):
                    if (detip in self.netmask):
                        theip = pair['addr']
                    else:
                        continue
                else:
                    theip = pair['addr']
                break  # we use the first IP found
        # theip can still be None at this point!
        self.set_current_value(theip)
        return theip

    def detect(self):
        return self._detect()



# -*- coding: utf-8 -*-

import logging
import re

import IPy
import requests

from .base import IPDetector

log = logging.getLogger(__name__)


def _get_ip_from_url(url, parser):
    log.debug("Querying IP address from '%s'", url)
    try:
        r = requests.get(url)
    except (requests.exceptions.RequestException) as exc:
        log.debug("webcheck failed for url '%s'", url, exc_info=exc)
        return None
    else:
        if r.status_code == 200:
            return parser(r.text)
        else:
            log.debug("Wrong http status code for '%s': %i", url, r.status_code)
    return None


def _parser_plain(text):
    try:
        return str(IPy.IP(text.strip()))
    except ValueError as exc:
        log.warn("Error parsing IP address '%s'", text, exc_info=exc)
        return None


def _parser_checkip(text):
    regex = re.compile("Current IP Address: (.*?)(<.*){0,1}$")
    for line in text.splitlines():
        matchObj = regex.search(line)
        if not matchObj is None:
            return str(IPy.IP(matchObj.group(1)))
    log.debug("Output '%s' could not be parsed", text)
    return None


def _parser_freedns_afraid(text):
    regex = re.compile("Detected IP : (.*?)(<.*){0,1}$")
    for line in text.splitlines():
        matchObj = regex.search(line)
        if not matchObj is None:
            return str(IPy.IP(matchObj.group(1)))
    log.debug("Output '%s' could not be parsed", text)
    return None


class IPDetector_WebCheck(IPDetector):
    """Class to detect an IP address as seen by an online web site that returns
    parsable output containing the IP address"""

    @staticmethod
    def getName():
        return "webcheck"

    def can_detect_offline(self):
        """Returns false, as this detector generates http traffic"""
        return False

    def detect(self):
        from random import choice
        urls = (
                ("http://checkip.dyndns.org/", _parser_checkip),
                ("http://checkip.eurodyndns.org/", _parser_checkip),
                ("http://dynamic.zoneedit.com/checkip.html", _parser_checkip),
                ("http://ipcheck.rehbein.net/", _parser_checkip),
                ("http://ip.dnsexit.com/", _parser_plain),
                ("http://freedns.afraid.org:8080/dynamic/check.php", _parser_freedns_afraid),
                ("http://icanhazip.com/", _parser_plain),
                ("http://ip.arix.com/", _parser_plain),
                ("http://ipv4.nsupdate.info/myip", _parser_plain),
                )
        theip = _get_ip_from_url(*choice(urls))
        if theip is None:
            log.info("Could not detect IP using webcheck! Offline?")
        self.set_current_value(theip)
        return theip



# -*- coding: utf-8 -*-

import logging
import warnings

from ..common.subject import Subject

log = logging.getLogger(__name__)


class IPDetector(Subject):
    """
    Base class for IP detectors. When implementing a new detector, it is
    usually best to just inherit from this class first.
    """
    def __init__(self, *args, **kwargs):
        super(IPDetector, self).__init__()

    def canDetectOffline(self):
        warnings.warn("canDetectOffline is deprecated; use "
               "can_detect_offline() instead", DeprecationWarning, stacklevel=2)
        return self.can_detect_offline()

    def can_detect_offline(self):
        """
        Must be overwritten in subclass. Return True if the IP detection
        does not generate any network traffic.
        """
        raise NotImplementedError("Abstract method, must be overridden")

    def get_old_value(self):
        try:
            self._oldvalue
        except AttributeError:
            return self.get_current_value()

    def getOldValue(self):
        warnings.warn("getOldValue is deprecated; use get_old_value() "
              "instead", DeprecationWarning, stacklevel=2)
        return self.get_old_value()

    def set_old_value(self, value):
        self._oldvalue = value

    def setOldValue(self, value):
        warnings.warn("setOldValue is deprecated; use set_old_value() "
              "instead", DeprecationWarning, stacklevel=2)
        self.set_old_value(value)

    def get_current_value(self, default=None):
        try:
            return self._currentvalue
        except AttributeError:
            return default

    def getCurrentValue(self, default=None):
        warnings.warn("getCurrentValue is deprecated; use get_current_value() "
              "instead", DeprecationWarning, stacklevel=2)
        return self.get_current_value(default)

    def set_current_value(self, value):
        if value != self.get_current_value():
            self._oldvalue = self.get_current_value(value)
            self._currentvalue = value
            log.debug("new IP set: %s", value)
        return value

    def setCurrentValue(self, value):
        warnings.warn("setCurrentValue is deprecated; use set_current_value() "
              "instead", DeprecationWarning, stacklevel=2)
        return self.set_current_value(value)

    def hasChanged(self):
        """Detect a state change with old and current value"""
        return self.get_old_value() != self.get_current_value()



# -*- coding: utf-8 -*-

import logging

import requests

from ..common.subject import Subject
from ..common.events import IP_UPDATE_SUCCESS, IP_UPDATE_ERROR

log = logging.getLogger(__name__)


class UpdateProtocol(Subject):
    """the base class for all update protocols"""

    _updateurl = None
    theip = None
    hostname = None  # this holds the desired dns hostname
    status = 0
    nochgcount = 0
    failcount = 0

    def __init__(self):
        self.updateurl = self._updateurl
        super(UpdateProtocol, self).__init__()

    def updateUrl(self):
        return self.updateurl

    def success(self):
        self.status = 0
        self.failcount = 0
        self.nochgcount = 0
        self.notify_observers(IP_UPDATE_SUCCESS, "Updated IP address of '%s' to %s" % (self.hostname, self.theip))

    def abuse(self):
        self.status = 1
        self.failcount = 0
        self.nochgcount = 0
        self.notify_observers(IP_UPDATE_ERROR, "This client is considered to be abusive for hostname '%s'" % (self.hostname))

    def nochg(self):
        self.status = 0
        self.failcount = 0
        self.nochgcount += 1

    def nohost(self):
        self.status = 1
        self.failcount += 1
        self.notify_observers(IP_UPDATE_ERROR, "Invalid/non-existant hostname: [%s]" % (self.hostname))

    def failure(self):
        self.status = 1
        self.failcount += 1
        self.notify_observers(IP_UPDATE_ERROR, "Service is failing")

    def notfqdn(self):
        self.status = 1
        self.failcount += 1
        self.notify_observers(IP_UPDATE_ERROR, "The provided hostname '%s' is not a valid hostname!" % (self.hostname))

    def protocol(self):
        params = {'myip': self.theip, 'hostname': self.hostname}
        r = requests.get(self.updateUrl(), params=params, auth=(self.userid, self.password), timeout=60)
        r.close()
        log.debug("status %i, %s", r.status_code, r.text)
        if r.status_code == 200:
            if r.text.startswith("good "):
                self.success()
                return self.theip
            elif r.text.startswith('nochg'):
                self.nochg()
                return self.theip
            elif r.text == 'nohost':
                self.nohost()
                return 'nohost'
            elif r.text == 'abuse':
                self.abuse()
                return 'abuse'
            elif r.text == '911':
                self.failure()
                return '911'
            elif r.text == 'notfqdn':
                self.notfqdn()
                return 'notfqdn'
            else:
                self.status = 1
                self.notify_observers(IP_UPDATE_ERROR, "Problem updating IP address of '%s' to %s: %s" % (self.hostname, self.theip, r.text))
                return r.text
        else:
            self.status = 1
            self.notify_observers(IP_UPDATE_ERROR, "Problem updating IP address of '%s' to %s: %s" % (self.hostname, self.theip, r.status_code))
            return 'invalid http status code: %s' % r.status_code



# -*- coding: utf-8 -*-

"""
module containing logic to run the dyndnsc cli program as a UNIX daemon
"""

import sys
import os


def daemonize(stdout=os.devnull, stderr=None, stdin=os.devnull,
              pidfile=None, startmsg='started with pid %s'):
    """
        This forks the current process into a daemon.
        The stdin, stdout, and stderr arguments are file names that
        will be opened and be used to replace the standard file descriptors
        in sys.stdin, sys.stdout, and sys.stderr.
        These arguments are optional and default to /dev/null.
        Note that stderr is opened unbuffered, so
        if it shares a file with stdout then interleaved output
        may not appear in the order that you expect.
    """
    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)  # Exit first parent.
    except OSError as e:
        sys.stderr.write("fork #1 failed: (%d) %s%s" % (e.errno, e.strerror, os.linesep))
        sys.exit(1)

    # Decouple from parent environment.
    os.chdir("/")
    os.umask(0)
    os.setsid()

    # interestingly enough, we MUST open STDOUT explicitly before we
    # fork the second time.
    # Otherwise, the duping of sys.stdout won't work,
    # and we will not be able to capture stdout
    sys.stdout.write('')

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)  # Exit second parent.
    except OSError as e:
        sys.stderr.write("fork #2 failed: (%d) %s%s" % (e.errno, e.strerror, os.linesep))
        sys.exit(1)

    # Open file descriptors and print start message
    if not stderr:
        stderr = stdout
    si = open(stdin, 'rb')
    so = open(stdout, 'w+b')
    se = open(stderr, 'w+b', 0)
    pid = str(os.getpid())
    sys.stderr.write("%s%s" % (startmsg, os.linesep) % pid)
    sys.stderr.flush()
    if pidfile:
        open(pidfile, 'w+b').write("%s%s" % (pid, os.linesep))

    # Redirect standard file descriptors.
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())



# -*- coding: utf-8 -*-

from .base import UpdateProtocol


class UpdateProtocolDummy(UpdateProtocol):

    _updateurl = "http://localhost.nonexistant/nic/update"

    def __init__(self, options=None):
        if options is None:
            self._options = {}
        else:
            self._options = options
        super(UpdateProtocolDummy, self).__init__()

    @staticmethod
    def configuration_key():
        return "dummy"

    def update(self, ip):
        return ip



# -*- coding: utf-8 -*-

import time
import logging
import warnings

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

log = logging.getLogger(__name__)


class DynDnsClient(object):
    """This class represents a client to the dynamic dns service."""
    def __init__(self, sleeptime=300):
        self.ipchangedetection_sleep = sleeptime  # check every n seconds if our IP changed
        self.forceipchangedetection_sleep = sleeptime * 5  # force check every n seconds if our IP changed
        self.lastcheck = None
        self.lastforce = None
        self.updaters = []
        self.dns = None
        self.detector = None
        self.status = 0
        log.debug("DynDnsClient instantiated")

    def add_updater(self, updater):
        self.updaters.append(updater)

    def setProtocolHandler(self, proto):
        warnings.warn("setProtocolHandler is deprecated; use add_updater() "
                      "instead", DeprecationWarning)
        self.add_updater(proto)

    def setDNSDetector(self, detector):
        self.dns = detector

    def setChangeDetector(self, detector):
        self.detector = detector

    def sync(self):
        """Forces a synchronization if there is a difference between the IP from
        DNS and the detector. This can be expensive, mostly depending on the
        detector, but also because updating the dynamic ip in itself is costly.
        Therefore, this method should usually only be called on startup or when
        the state changes.
        """
        if self.dns.detect() != self.detector.detect():
            detected_ip = self.detector.get_current_value()
            if detected_ip is None:
                log.debug("DNS is out of sync, but we don't know what to update it to (detector returned None)")
                self.status = 2
                # we don't have a value to set it to, so don't update! Still shouldn't happen though
            else:
                log.info("Current dns IP '%s' does not match detected IP '%s', updating", self.dns.get_current_value(), detected_ip)
                # TODO: perform some kind of proxy chaining here?
                for ipupdater in self.updaters:
                    status = ipupdater.update(detected_ip)
                self.status = status
        else:
            self.status = 0
            log.debug("Nothing to do, dns '%s' equals detection '%s'",
                      self.dns.get_current_value(),
                      self.detector.get_current_value())

    def stateHasChanged(self):
        """Detects a change either in the offline detector or a
        difference between the real DNS value and what the online
        detector last got.
        This is efficient, since it only generates minimal dns traffic
        for online detectors and no traffic at all for offline detectors.

        @return: boolean
        """
        self.lastcheck = time.time()
        # prefer offline state change detection:
        if self.detector.can_detect_offline():
            self.detector.detect()
        elif not self.dns.detect() == self.detector.get_current_value():  # query current dns, and only detect if there's no match
            # this produces traffic, but probably less traffic overall than the detector
            self.detector.detect()
        if self.detector.hasChanged():
            log.debug("detector changed")
            return True
        elif self.dns.hasChanged():
            log.debug("dns changed")
            return True
        else:
            return False

    def needsCheck(self):
        """This checks if the planned time between checks has elapsed.
        When this time has elapsed, a state change check through
        stateHasChanged() should be performed and eventually a sync().
        """
        if self.lastcheck is None:
            return True
        else:
            return time.time() - self.lastcheck >= self.ipchangedetection_sleep

    def needsForcedCheck(self):
        """This checks if self.forceipchangedetection_sleep between checks has
        elapsed. When this time has elapsed, a sync() should be performed, no
        matter what stateHasChanged() says. This is really just a safety thing
        to enforce consistency in case the state gets messed up.
        """
        if self.lastforce is None:
            self.lastforce = time.time()
        elapsed = time.time() - self.lastforce
        if (elapsed < self.forceipchangedetection_sleep):
            return False
        return True

    def check(self):
        if self.needsCheck():
            log.debug("needs a check according to ipchangedetection_sleep (%s sec)", self.ipchangedetection_sleep)
            if self.stateHasChanged():
                log.debug("state changed, syncing...")
                self.sync()
            elif self.needsForcedCheck():
                log.debug("forcing sync after %s seconds",
                          self.forceipchangedetection_sleep)
                self.lastforce = time.time()
                self.sync()
            else:
                # nothing to be done
                pass

    def loop(self):
        """Blocking endless loop with built-in sleeping between checks and
        updates."""
        while True:
            self.check()
            time.sleep(self.ipchangedetection_sleep)


def getDynDnsClientForConfig(config):
    """Factory method to instantiate and initialize a complete and working
    dyndns client

    @param config: a dictionary with configuration pairs
    """
    if config is None:
        return None
    if not 'hostname' in config:
        log.warn("No hostname configured")
        return None
    from .detector import dns
    dns_detector = dns.IPDetector_DNS(config['hostname'])
    from dyndnsc.updater import getUpdaterClass
    try:
        klass = getUpdaterClass(config['protocol'])
    except KeyError:
        log.warn("Invalid update protocol: '%s'", config['protocol'])
        return None
    try:
        ip_updater = klass(config)
    except (AssertionError, KeyError) as exc:
        log.warn("Invalid update protocol configuration: '%s'", repr(config),
                 exc_info=exc)
        return None

    dyndnsclient = DynDnsClient(sleeptime=config['sleeptime'])
    dyndnsclient.add_updater(ip_updater)
    dyndnsclient.setDNSDetector(dns_detector)

    # allow config['method'] to be a list or a comma-separated string:
    if type(config['method']) != list:
        dummy = config['method'].split(',')
    else:
        dummy = config['method']
    method = dummy[0]
    if len(dummy) > 1:
        method_optlist = dummy[1:]
    else:
        method_optlist = []
    from .detector import manager
    try:
        klass = manager.get_detector_class(method)
    except (KeyError) as exc:
        log.warn("Invalid change detector configuration: '%s'", method,
                 exc_info=exc)
        return None

    # make a dictionary from method_optlist:
    opts = {}
    colon = ":"
    for opt in method_optlist:
        # options are key value pairs, separated by a colon ":"
        # allow white-spaces in input, but strip them here:
        option, dummysep, value = opt.partition(colon)
        option = option.strip()
        if option in opts:
            log.warn("Option '%s' specified more than once, using '%s'.",
                     option, value)
        opts[option] = value.strip()
    try:
        dyndnsclient.setChangeDetector(klass(opts))
    except KeyError as exc:
        log.warn("Invalid change detector parameters: '%s'", opts, exc_info=exc)
        return None

    return dyndnsclient



# -*- coding: utf-8 -*-

import unittest
import logging

import dyndnsc

logging.basicConfig(level=logging.DEBUG)


class DynDnscTestCases(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def test_dummy(self):
        config = {}
        config['hostname'] = "dyndnsc.no-ip.biz"
        config['userid'] = "example"
        config['password'] = "foobar"
        config['protocol'] = "dummy"
        config['method'] = "random"
        config['sleeptime'] = 60
        dyndnsclient = dyndnsc.getDynDnsClientForConfig(config)
        self.assertTrue(dyndnsclient.needsCheck())
        dyndnsclient.needsForcedCheck()
        dyndnsclient.check()
        dyndnsclient.sync()
        dyndnsclient.stateHasChanged()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

classifiers = [line.strip() for line in """
Development Status :: 3 - Alpha
Intended Audience :: Developers
License :: DFSG approved
License :: OSI Approved
License :: OSI Approved :: MIT License
Topic :: Internet :: Name Service (DNS)
Topic :: Software Development :: Libraries :: Python Modules
Environment :: Console
Natural Language :: English
Operating System :: MacOS :: MacOS X
Operating System :: POSIX :: Linux
Operating System :: POSIX :: BSD :: FreeBSD
Programming Language :: Python
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.2
Programming Language :: Python :: 3.3
""".splitlines() if len(line) > 0]

install_requires = ["requests", "IPy>=0.56"]
if sys.version_info >= (3, 0):
    install_requires.append("netifaces-py3==0.8")
else:
    install_requires.append("netifaces>=0.4")
    # work around python issue http://bugs.python.org/issue15881
    # affects only python2 when using multiprocessing and if nose is installed
    import multiprocessing

if sys.version_info < (3, 2):
    install_requires.append("argparse")
if sys.version_info < (2, 7):  # continue support for python 2.6
    install_requires.append("importlib")

setup(name="dyndnsc",
      packages=["dyndnsc", "dyndnsc.common", "dyndnsc.detector",
                "dyndnsc.updater", "dyndnsc.tests"],
      version="0.3.2",
      author="Paul Kremer",
      author_email="@".join(("paul", "spurious.biz")),  # avoid spam,
      license="MIT License",
      description="dynamic dns update client package that tries to be "
            "extensible, re-usable and efficient on network resources",
      long_description=(open("README.rst", "r").read() + "\n\n" +
                        open("CHANGELOG.rst", "r").read()),
      url="https://github.com/infothrill/python-dyndnsc",
      install_requires=install_requires,
      entry_points=("""
                      [console_scripts]
                      dyndnsc=dyndnsc.cli:main
                      """),
      classifiers=classifiers,
      test_suite='dyndnsc.tests',
      tests_require=['bottle==0.11.6'],
      )



# -*- coding: utf-8 -*-

import logging
import random

import IPy

from .base import IPDetector

log = logging.getLogger(__name__)


class RandomIPGenerator(object):
    def __init__(self, maxRandomTries=None):
        self.maxRandomTries = maxRandomTries

        # Reserved list from http://www.iana.org/assignments/ipv4-address-space
        # (dated  2010-02-22)
        self._reserved_netmasks = frozenset([
                "0.0.0.0/8",
                "5.0.0.0/8",
                "10.0.0.0/8",
                "23.0.0.0/8",
                "31.0.0.0/8",
                "36.0.0.0/8",
                "39.0.0.0/8",
                "42.0.0.0/8",
                "127.0.0.0/8",
                "169.254.0.0/16",
                "172.16.0.0/12",
                "192.168.0.0/16",
                "224.0.0.0/3",
                "240.0.0.0/8"
                ])

    def isReservedIP(self, ip):
        """Check if the given ip address is in a reserved ipv4 address space

        @param ip: IPy ip address
        @return: boolean
        """
        for res in self._reserved_netmasks:
            if ip in IPy.IP(res):
                return True
        return False

    def randomIP(self):
        """Return a randomly generated IPv4 address that is not in a reserved
        ipv4 address space

        @return: IPy ip address
        """
        randomip = IPy.IP("%i.%i.%i.%i" % (random.randint(1, 254),
                                           random.randint(1, 254),
                                           random.randint(1, 254),
                                           random.randint(1, 254)))
        while self.isReservedIP(randomip):
            randomip = IPy.IP("%i.%i.%i.%i" % (random.randint(1, 254),
                                               random.randint(1, 254),
                                               random.randint(1, 254),
                                               random.randint(1, 254)))
        return randomip

    def next(self):
        return self.__next__()

    def __next__(self):
        """Generator that returns randomly generated IPv4 addresses that are
        not in a reserved ipv4 address space until we hit self.maxRandomTries

        @return: IPy ip address
        """
        if self.maxRandomTries is None or self.maxRandomTries > 0:
            generate = True
        c = 0
        while generate:
            if not self.maxRandomTries is None:
                c += 1
            yield self.randomIP()
            if not self.maxRandomTries is None and c < self.maxRandomTries:
                generate = False

        raise StopIteration

    def __iter__(self):
        """Iterator for this class. See method next()"""
        return next(self)


class IPDetector_Random(IPDetector):
    """For testing: detect randomly generated IP addresses"""
    def __init__(self, options=None):
        super(IPDetector_Random, self).__init__()
        self.rips = RandomIPGenerator()

    @staticmethod
    def getName():
        return "random"

    def can_detect_offline(self):
        """Returns True"""
        return True

    def detect(self):
        for theip in self.rips:
            log.debug('detected %s', str(theip))
            self.set_current_value(str(theip))
            return str(theip)



# -*- coding: utf-8 -*-

import logging

from .iface import IPDetector_Iface

log = logging.getLogger(__name__)


class IPDetector_Teredo(IPDetector_Iface):
    """IPDetector to detect a Teredo ipv6 address of a local interface.
    Bits 0 to 31 of the ipv6 address are set to the Teredo prefix (normally
    2001:0000::/32).
    This detector only checks the first 16 bits!
    See http://en.wikipedia.org/wiki/Teredo_tunneling for more information on
    Teredo.

    Inherits IPDetector_Iface and sets default options only
    """
    def __init__(self, options=None):
        """
        Constructor
        @param options: dictionary
        """
        if options is None:
            options = {}
        super(IPDetector_Teredo, self).__init__(options)
        self.opts = {'iface': 'tun0', 'family': "INET6",
                     "netmask": "2001:0000::/32"}
        for k in options.keys():
            log.debug("%s explicitly got option: %s -> %s",
                      self.__class__.__name__, k, options[k])
            self.opts[k] = options[k]

    @staticmethod
    def getName():
        return "teredo"



# -*- coding: utf-8 -*-

"""
Example CLI program
"""

import sys
import pkg_resources
import logging

from dyndnsc import getDynDnsClientForConfig
from dyndnsc.daemon import daemonize


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--daemon", dest="daemon",
                        help="go into daemon mode (implies --loop)",
                        action="store_true", default=False)
    parser.add_argument("--debug", dest="debug",
                        help="increase logging level to DEBUG",
                        action="store_true", default=False)
    parser.add_argument("--hostname", dest="hostname",
                        help="hostname to update", default=None)
    parser.add_argument("--key", dest="key",
                        help="your authentication key", default=None)
    parser.add_argument("--userid", dest="userid",
                        help="your userid", default=None)
    parser.add_argument("--password", dest="password",
                        help="your password", default=None)
    parser.add_argument("--protocol", dest="protocol",
                        help="protocol to use for updating IP (default dyndns)",
                        default='dyndns')
    parser.add_argument("--method", dest="method",
                        help="method for detecting your IP (default webcheck)",
                        default='webcheck')
    parser.add_argument("--loop", dest="loop",
                        help="loop forever (default is to update once)",
                        action="store_true", default=False)
    parser.add_argument("--sleeptime", dest="sleeptime",
                        help="how long to sleep between checks in seconds",
                        default=300)
    parser.add_argument("--version", dest="version",
                        help="show version and exit",
                        action="store_true", default=False)
    args = parser.parse_args()

    if args.version:
        print("dyndnsc %s" % pkg_resources.get_distribution("dyndnsc").version)  # pylint: disable=E1103
        return 0

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(level=level,
                        format='%(asctime)s %(levelname)s %(message)s')
    # silence 'requests' logging
    requests_log = logging.getLogger("requests")
    requests_log.setLevel(logging.WARNING)

    if args.hostname is None:
        parser.error("Please specify a hostname using --hostname")

    config = {}
    config['hostname'] = args.hostname
    config['key'] = args.key
    config['userid'] = args.userid
    config['password'] = args.password
    config['protocol'] = args.protocol
    config['method'] = args.method
    config['sleeptime'] = int(args.sleeptime)

    # done with command line options, bring on the dancing girls
    dyndnsclient = getDynDnsClientForConfig(config)
    if dyndnsclient is None:
        return 1
    # do an initial synchronization, before going into endless loop:
    dyndnsclient.sync()

    if args.daemon:
        daemonize()  # fork into background
        args.loop = True

    if args.loop:
        dyndnsclient.loop()
    else:
        dyndnsclient.check()

    return 0

if __name__ == '__main__':
    sys.exit(main())
